######################################################################
#:CQCC::Parm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

=head1 NAME

CQCC::Parm - Class used to manage class parameters

=head1 CLASS DESCRIPTION

This class manages all class parameters and facilitates setting them 
from configuration files, environment variables, command line options,
and other sources. 

The class methods are not invoked by callers directly but through a
series of convenience functions which handle Parm objects through a 
set of hash lists. 

DefineParm() is called by the owning class to describe the parameter
and its limitations and sources.

SetConfigParm() , SetCmdParm() , and SetIntParm() are functions which 
will remember a "preference" for the parameter's value on their own
separate lists. This allows parameter settings to be made before the
parameter has been defined. 

GetParm() "evaluates" the parameter by seeing if there is a source for
a non-default value that fits the boundary conditions (multiple
available sources are evaluated in priority order); if not it will
return the default value if one is defined. Once the parameter has
been evaluated its value is stored so that later accesses to it will
not require re-evaluation. 

ForceParm() can be used to force overrides by clearing the current
setting, storing an Internal parameter value (highest priority) and
reevaluating the parameter.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC:Parm
######################################################################

package CQCC::Parm;
use Exporter;
@ISA = ('Exporter');
@EXPORT = qw(&DefineParm &GetParm &SetConfigParm &GetParmNameList
             &SetCmdParm &SetIntParm &EvalParms &PrintParms
             &ParmString &ParmParse &SetParmDefaults
             &ForceParm &ParmDefined &GetParmHandle &UpdateParmHandle);

use strict;

sub Version { return '2.1' };

######################################################################
#::Class Globals 
######################################################################

my %ParmList = ();       # Hash holding defined parm objects
my %ParmValue = ();      # Hash holding approved values
my %ParmValueConverted = (); # Hash holding converted values
my %ParmConfig = ();     # Hash holding admin provided values
my %ParmCmdLine = ();    # Hash holding command line values (-p n v)
my %ParmInternal = ();   # Hash holding internal values

my @ParmDefaults = ();   # Defaults used in creating Parameters

my %parmHandleEncoding = ();
my %parmHandleValues = ();

my ($_ConvertFunction,       # IOConvertString caller 
    $_MessageFunction        # IOMessage function if available
   );

######################################################################
#::function ClassInit
######################################################################

#sub ClassInit {}
#&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This method will create samples of each type of parameter then
evaluate them, to ensure that range checking and other aspects
of evaluation works. 

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;

    # The _TEST value is used only as a handle for the PrintParms 
    # operation so we don't bring in all the other defined parameters
    # during a classtest.

    DefineParm(NAME=>"Test1_EV",
               PCLASS=>"CQCC::Parm",
               SRC_EV=>1, _TEST=>1,
               DEFAULT=>15,
               RANGE_MIN=>10,
               RANGE_MAX=>100);

    DefineParm(NAME=>"Test1_CMDLINE",
               PCLASS=>"CQCC::Parm",
               SRC_CMDLINE=>1, _TEST=>1,
               DEFAULT=>35,
               RANGE_MIN=>10,
               RANGE_MAX=>100);

    DefineParm(NAME=>"Test1_CONFIG",
               PCLASS=>"CQCC::Parm",
               SRC_CFG=>1, _TEST=>1,
               DEFAULT=>50,
               RANGE_MIN=>10,
               RANGE_MAX=>100);

    DefineParm(NAME=>"Test1_LIST",
               PCLASS=>"CQCC::Parm",
               TYPE=>"LIST", _TEST=>1,
               SRC_CFG=>1,
               );

    DefineParm(NAME=>"Test1_INTERNAL",
               PCLASS=>"CQCC::Parm",
               SRC_INT=>1, _TEST=>1,
               );

    $ENV{"Test1_EV"} = 3;
    SetConfigParm("Test1_CONFIG", 150);
    SetCmdParm("Test1_CMDLINE", 2);
    SetConfigParm("Test1_LIST", [1,2,3]);
    SetIntParm("Test1_INTERNAL", 5);

    print "Expected results:\n";
    print "Test1_CMDLINE - 2 < 10-100 range, set to 35 default\n";
    print "Test1_CONFIG  - 150 > 10-100 range, set to 50 default\n";
    print "Test1_EV - 3 < 10-100 range, set to 15 default\n";
    print "\n\n";
    EvalParms();
    my $a = PrintParms("_TEST");  # Should self-print results

    my $str1 = &ParmString("Test1_CONFIG");
    print "Should be Test1_CONFIG=50\n";
    print $str1, "\n";
  
    ForceParm("Test1_CONFIG", 10);
    ParmParse($str1);
    my $str2 = &ParmString("Test1_CONFIG");
    if ($str1 ne $str2) { 
        print "Error: ParmParse did not reset Test1_CONFIG\n";
    } else {
        print "ParmParse successfully restored value from string\n";
    } 
    return 0;
}

######################################################################
#::function DefineParm

=over 6

=item C<DefineParm(keys)>

This function creates a new parm object and stores it into the Parm
array to control how GetParm will handle evaluating it. It does NOT
evaluate the function, in order to allow other events to set preferred
values before it is needed. 

=back

=cut

######################################################################

sub DefineParm { 
    my @args = @ParmDefaults;
    push @args, @_;
    my $new_parm = new CQCC::Parm(@args);
    if ($ParmList{$new_parm->{NAME}}) {
        &_Message("WARN", ["PM_E_TWO_PARMS", $new_parm->{NAME}]);
    } else {
        $ParmList{$new_parm->{NAME}} = $new_parm;
    }
}

######################################################################
#::function EvalParms

=over 6

=item C<EvalParms()>

This function will force evaluation of all parms - otherwise parms are
evaluated only as needed in "lazy evaluation" mode.  PrintParms will
effectively accomplish the same thing.

=back

=cut 

######################################################################

sub EvalParms {
    my ($k, $v);

    foreach $k (sort keys(%ParmList)) {
        $v = &GetParm($k);
    }
}

######################################################################
#::function ForceParm

=over 6

=item C<ForceParm($name,$value,@options)>

This internal function forces parameter values as internal overrides.
It will clear any existing value, set its own value as an internal
parameter, and call GetParm to reevaluate the parameter.  This will
ensure all normal checks are still performed but the provided value
will override all other sources.

The options list provides optional parameters, currently as follows:

* USER_CHANGED => 1 - Indicate user interactively changed value

=back

=cut

######################################################################

sub ForceParm { 
    my ($p_name, $p_value,%options) = @_;
    $ParmValue{$p_name} = undef;   # Clear existing value, reevaluate
    $ParmValueConverted{$p_name} = undef; # Clear converted value
    &SetIntParm($p_name, $p_value);
    if ($options{"USER_CHANGED"} eq 1) {
        my $parm = $ParmList{$p_name};
        if ($parm->{SRC_USER_PREF}) {
            $parm->{USER_CHANGED_PREF} = 1;
        }
        # Retrieve using GetParmNameList{USER_CHANGED_PREF}
        # to find out if user changed any preferences
        # and GetParmNameList{SRC_USER_PREF} to get all prefs to save
    }

    &UpdateParmHandle($p_name);

    return &GetParm($p_name);
} 

######################################################################
#::function GetParm

=over 6

=item C<GetParm($name)>

This function will lookup the Parm for "$name". If it already
has been evaluated then its value will be returned immediately.
Otherwise its definition will be checked for acceptable sources
and value ranges. Once evaluated the value will be stored for
future use and is returned to the original caller. 

=back

=cut

######################################################################

sub GetParm {
    my $p_name = shift;

    my $p_value = $ParmValueConverted{$p_name};

    # Simplest case - return a previously evaluated value
    if (defined $p_value) { 
        return $p_value;
    } 
    
    ##############################
    # Confirm the parm object has already been defined
    my $failed_test = 0;
    my $self = $ParmList{$p_name};

    if (! defined $self) {
        &_Message("ERROR", ["PM_E_UNKNOWN_PARM", $p_name]);
    } 

    ##############################
    # Preset value to the default in case proposed value fails. 

    $ParmValue{$p_name} = $self->{DEFAULT};

    # Obtain the highest priority proposed value 
    # Always take an internal value if it is available as an override
    if (#defined $self->{SRC_INT} && 
           defined $ParmInternal{$p_name}) {
       $p_value = $ParmInternal{$p_name};
    } elsif (defined $self->{SRC_EV} && defined $ENV{$p_name}) {
       $p_value = $ENV{$p_name};            
    } elsif (defined $self->{SRC_CMD} && 
           defined $ParmCmdLine{$p_name}) {
       $p_value = $ParmCmdLine{$p_name};
    } elsif (defined $self->{SRC_CFG} && 
             defined $ParmConfig{$p_name}) {
       $p_value = $ParmConfig{$p_name};
    # If no alternatives provided, use default, assume its ok
    } else {
       return &_ConvertedParmValue($p_name);
    }

    ##############################
    # Evaluate the proposed value, applying ALL tests

    if (defined $self->{RANGE_MIN}) {
        if ($p_value < $self->{RANGE_MIN}) {
            &_Message("WARN", ["PM_E_PARM_RANGE", $p_name, $p_value,
               ($self->{RANGE_MIN} ? 
                     $self->{RANGE_MIN} : "(no min)"),
               ($self->{RANGE_MAX} ? 
                     $self->{RANGE_MAX} : "(no max)")]);
            $failed_test = 1;
        }
    }

    if (defined $self->{RANGE_MAX}) {
        if ($p_value > $self->{RANGE_MAX}) {
            &_Message("WARN", ["PM_E_PARM_RANGE", $p_name, $p_value,
               ($self->{RANGE_MIN} ? 
                     $self->{RANGE_MIN} : "(no min)"),
               ($self->{RANGE_MAX} ? 
                     $self->{RANGE_MAX} : "(no max)")]);
            $failed_test = 1;
        }
    }

    if (defined $self->{RANGE_LIST}) {
        my @p_list = @{ $self->{RANGE_LIST} };
        my $found = 0;
        for (@p_list) { 
            if ($p_value eq $_) { 
                $found = 1;
            } 
        } 
        if (! $found) {
            &_Message("WARN", ["PM_E_PARM_LIST", 
               $p_name, $p_value, join(", ", @p_list)]);
            $failed_test = 1;
        }
    }

    if (! $failed_test) { 
       $ParmValue{$p_name} = $p_value;
    } 

    return &_ConvertedParmValue($p_name);
}
 
######################################################################
# Conversion routine: 
# If value is undefined or a converted value exists, return that. 
# Otherwise try to convert the value
# If conversion handler is set then we can safely save the converted
# value for faster access in future references until it changes
# (ForceParm will clear the converted value before reevaluation).

sub _ConvertedParmValue {
    my ($p_name) = @_;
    my $p_value = $ParmValue{$p_name};
        
    # If ParmValue is undefined, no need to convert
    if (! defined $p_value) {
        return $p_value;
    } 
    elsif ($ParmValueConverted{$p_name}) {
        return $ParmValueConverted{$p_name};
    }
    # Need to convert the value 
    # Assume value may be utf8 and convert it to core
    my ($status,$c_value) = 
        &ParmIOConvertString($p_value,"UTF8_CODEPAGE","CORE_CODEPAGE");
    if ($status) {
        &ParmIOMessage("FATAL", ["PM_E_CONVERSION_ERROR",$p_name,$p_value]);
    }
    # If ConvertFunction was registered its safe to save converted value
    elsif ($_ConvertFunction) {
        $ParmValueConverted{$p_name} = $c_value;
    }
    return $c_value;
}

######################################################################
#::function GetParmHandle

=over 6

=item C<GetParmHandle($name)>

This internal method gets a reference to the value cell for the named
parameter. It is only intended for use by GUI widgets that are interacting
with the parameter. 

=back

=cut

######################################################################

sub GetParmHandle {
    my ($p_name, $encoding) = @_;
 
    if ($encoding) {
        $parmHandleEncoding{$p_name} = $encoding;
        UpdateParmHandle($p_name);
        return \$parmHandleValues{$p_name};
    }else{
    my $value = &GetParm($p_name);
    return \$ParmValue{$p_name};
}
}

######################################################################
#::function UpdateParmHandle

=over 6

=item C<UpdateParmHandle(name)>

This function will update the named parameter value by converting the
original value in the CORE encoding. This is only needed for parameters
that are stored in one encoding and displayed in another.

=back

=cut

######################################################################

sub UpdateParmHandle {
    my ($name) = @_;

    my $encoding = $parmHandleEncoding{$name};
    if ($encoding) {
        my ($status,$value) = 
            &ParmIOConvertString(&GetParm($name),"CORE_CODEPAGE",$encoding);
        if ($status) {
            &ParmIOMessage("FATAL", ["PM_E_CONVERSION_ERROR",$name,GetParm($name)]);
        }         
        $parmHandleValues{$name} = $value;
    } 
}
 
######################################################################
#::function GetParmNameList

=over 6

=item C<GetParmNameList($src)>

This function will lookup the Parms that have the requested source
and return a list of names that match. If no source is provided 
then the entire Parm list is returned. 

=back

=cut

######################################################################

sub GetParmNameList {
    my ($src) = @_;
    my @parmList; 

    for (keys %ParmList) { 
        my $parm = $ParmList{$_};
        if (! defined $src || $parm->{$src}) { 
            push @parmList, $parm->{NAME};
        }
    } 

    return @parmList;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::Parm()>

This method defines a new parm and add it to the parm hash list. 
Nothing is evaluated at definition time but the information is
used to control the later evaluation when the variable is needed.

A Parm can be preset from one or more sources including, in order of
priority, environment variables (SRC_EV=>1), command line options
(SRC_CMD=>1), configuration settings (SRC_CFG=>1) or internal calls
(SRC_INT=>1). Which sources are valid depends on the definition of the
parameter. This allows complete control over which sources are allowed
for each parameter. An additional source (SRC_CACHE) is supported
but the parameters must also have an additional source such as 
SRC_INT.

A Parm can have a DEFAULT value set when it is defined and can have
rudimentary value checking done by setting RANGE_MIN, RANGE_MAX or
RANGE_LIST which will apply their respective tests when the Parm is
being evaluated for the first time by GetParm(). 

Each Parm is assumed to have a message catalog entry that describes
its use in appropriate detail for end user consumption. 

The new() method is not called directly but is invoked by the 
convenience function DefineParm(). 

Errors: 

(1) Parm with same name already exists.

(2) Need to add error checks for complete enough information

=back

=cut

######################################################################

sub new { 
   my $proto = shift;

   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self = {
       # Required instance data
       NAME => undef,       # Parm Name
       CLASSNAME => "CQCC::Parm", 
       DEFAULT => undef,    # Default value if none provided
       TYPE => "SCALAR",    # SCALAR, LIST 
       PCLASS => undef,     # Class name that owns the parm
       
       # Optional instance data 
       EXIT_IF_FAIL => 1,   # Exit if provided value doesn't pass 
                            #     limits, if 0 only provide warning.

       # Source - more than one can be used, if none then default used
       # SRC_EV=>1,         # Enables use as an environment variable
       # SRC_CMD=>1,        # Enables use as an command line option
       # SRC_CFG=>1,        # Enables use as an config variable
       # SRC_INT=>1,        # Enables use as an internal variable
       # SRC_CACHE=>1,      # Flags cached parameters
       # SRC_USER_PREF=>undef, # Enables saving value in session wide cache

       # USER_CHANGED=>undef,   # Flags parameter as user originated change

       # HELPTAG => undef,  # String we can use to look up description
                            #   If not set, assumed same as NAME
       # RANGE_MIN => undef,  # Minimum value
       # RANGE_MAX => undef,  # Maximum value
       # RANGE_LIST => undef, # List of allowed values

       @_,                  # Override previous attributes
       @_,                  # Extra time for ccperl
   };

   # Add test that all required values (Name and defval)
   
   bless($self, $class);
   return $self;
}

######################################################################
#::function ParmDefined

=over 6

=item C<ParmDefined($name, { $source })>

This function will lookup the Parm for "$name" and return 1 if
it is found, 0 if it doesn't. This permits a ClassInit function to
see if it has already been called. 

The optional $source argument also checks if a given data source
provides a value, i.e. if "CONFIG", the method will only return 1
if the parameter exists and SetConfigParm has already been called 
for the parameter to set a configuration value. Permitted values 
include "CONFIG", "CMDLINE", and "ENV".

=back

=cut

######################################################################

sub ParmDefined {
    my ($p_name, $p_source) = @_;

    if ($ParmList{$p_name}) {
        if ($p_source) { 
            if (uc($p_source) eq "CONFIG") {
        if ($ParmConfig{$p_name}) {
            return 1;
                }
            } elsif (uc($p_source) eq "CMDLINE") {
        if ($ParmCmdLine{$p_name}) {
            return 1;
                }
            } elsif (uc($p_source) eq "ENV") {
                if ($ENV{$p_name}) {
                    return 1;
                }
            }
            return 0;
        }
        return 1;
    }
    return 0;
}

######################################################################
#::function ParmIOConvertString

=over 6

=item C<($status,value) = ParmIOConvertString(source, src_encoding, dst_encoding)>

This function will use the registered conversion function if available 
to convert the string into the destination codepage, returning an error 
object if appropriate.

=back

=cut

######################################################################

sub ParmIOConvertString {
    my ($src_input, $src_encoding, $dst_encoding) = @_;
    
    if ($_ConvertFunction) {
        return &$_ConvertFunction($src_input, $src_encoding, $dst_encoding);        
    }
    #Note: Convert function returns list references as lists
    if (ref($src_input) eq "ARRAY") { 
        return (undef, @{ $src_input });
    }
    return (undef, $src_input);
}

sub SetConvertHandler {
    my ($handler) = @_;
    $_ConvertFunction = $handler;
}

######################################################################
#::function ParmIOMessage

=over 6

=item C<ParmIOMessage($msgTag)>

This function handles internal error messages by using the IOMessage function
handle provided by IOBase or by using print.

=back

=cut

######################################################################

sub ParmIOMessage {
    my ($type, $msgTag, %options) = @_;
    if ($_MessageFunction) {
       return &$_MessageFunction(@_);
    }
    print "$type: $msgTag\n";
} 

sub SetMessageHandler {
    my ($handler) = @_;
    $_MessageFunction = $handler;
}

######################################################################
#::function ParmParse

=over 6

=item C<ParmParse($string)>

This function will parse a string of the form "parmname=value"
that has been used to store the parameter in a file or read back
a value from the user. The complement function to generate this string
is ParmString($name);

=back

=cut

######################################################################

sub ParmParse{
    my $p_string = shift;

    if ($p_string =~ m/(\S*)\=(.*)/) {
        my $p_name = $1;
        my $p_value = $2;
        my $p_parm = $ParmList{$p_name};
        if ($p_parm) {
           if ($p_parm->{TYPE} eq "LIST") { 
               my @p_list = split("!!",$p_value);        
               &ForceParm($p_name, \@p_list);
           } else { 
               &ForceParm($p_name, $p_value);
           }
        } else { 
           &_Message("ERROR", ["PM_E_UNKNOWN_PARM", $p_name]);
        }
    } else {
#<DJS> Invalid pattern
    } 
}

######################################################################
#::function ParmString

=over 6

=item C<ParmString($name)>

This function will return a string of the form "parmname=value"
that can be used to store the parameter into a file or display it
for debugging purposes. The complement function to read back the
value is ParseParm($string). 

=back

=cut

######################################################################

sub ParmString{
    my $p_name = shift;
    my $p_value = &GetParm($p_name);

    if (! defined $p_value) {
        return;
    }
    return "$p_name=$p_value";
} 

######################################################################
#::function PrintParms

=over 6

=item C<PrintParms( {source} )>

This function will print the evaluated values of all defined parms.
If the optional source argument is provided, only parameters that
use that source will be shown, i.e. "SRC_EV" will only show
parameters that are accessible through environment variables. 

If a list result is expected, it will return the list of printable
values rather than printing them directly. 
=back

=cut 

######################################################################

sub PrintParms {
    my $source = shift;

    my $k;
    my @results;
    
    foreach $k (sort keys(%ParmList)) {
        my $self = $ParmList{$k};
        if ((! defined $source) || ($self->{$source} == 1)) {
            if ($self->{TYPE} eq "LIST") {
                my $v = &GetParm($k);
                my @va = (defined $v ? @{ $v } : []);
                push @results, "Parm $k = [ @va ]";
            } else {
                my $v = &GetParm($k);
                push @results, "Parm $k = $v";
            }
        }
    }
    if (wantarray()) {
        return @results;
    }
    else {
    print join("\n",@results),"\n";
        return;
    } 
}

######################################################################
#::function SetConfigParm

=over 6

=item C<SetConfigParm($name,$value)>

This function is used by administrators in configuration files to set
preferences for class parameters.

It stores a parameter value "preference" that will be used only if it
is higher priority than other sources. It may be called before
DefineParm() has defined the parameter but must be called before
GetParm() has a chance to evaluate other options.

=back

=cut

######################################################################

sub SetConfigParm { 
    my ($p_name, $p_value) = @_;
    $ParmConfig{$p_name} = $p_value;
} 

######################################################################
#::function SetCmdParm

=over 6

=item C<SetCmdParm($name,$value)>

This function is used by the trigger main loop as it evaluates its
command line arguments of the form -p name,value to allow the command
line to override other sources.

It stores a parameter value "preference" that will be used only if it
is higher priority than other sources. It may be called before
DefineParm() has defined the parameter but must be called before
GetParm() has a chance to evaluate other options.

=back

=cut

######################################################################

sub SetCmdParm { 
    my ($p_name, $p_value) = @_;
    $ParmCmdLine{$p_name} = $p_value;
} 

######################################################################
#::function SetIntParm

=over 6

=item C<SetIntParm($name,$value)>

This function is used by the integration to force the override of
other sources in some rare cases, such as allowing -test mode to force
debugging to be on.

It stores a parameter value "preference" that will be used only if it
is higher priority than other sources. It may be called before
DefineParm() has defined the parameter but must be called before
GetParm() has a chance to evaluate other options.

=back

=cut

######################################################################

sub SetIntParm { 
    my ($p_name, $p_value) = @_;
    $ParmInternal{$p_name} = $p_value;
} 

######################################################################
#::function SetParmDefaults

=over 6

=item C<SetParmDefaults(keys)>

Convenience function to set defaults for DefineParm calls for a set of
parameters. Should be used before to set defaults and after to reset
to () to avoid side effects in other areas.  Not required to use
DefineParm.

=back

=cut

######################################################################

sub SetParmDefaults {
    @ParmDefaults = @_;
}

######################################################################
#::function _Message
# Internal function used to handle errors for now. Because of Perl
# initialization behavior while loading the application, it is 
# unable to use IOBase without getting into circular references
# between multiple packages.
# TODO: Combine this with ParmIOMessage instead
######################################################################

sub _Message {
    my ($type, $msgtag) = @_;

    my ($tag, @args) = @{ $msgtag };
    my %messageTags = (
      "PM_E_PARM_RANGE" => 
          "Parameter %s value %s is outside range (%s to %s)",

      "PM_E_PARM_LIST" =>
         "Parameter %s value %s is not on the accepted list %s",

      "PM_E_TWO_PARMS" =>
           "Two parameters are using the same name or being " . 
           "defined twice: %s",

      "PM_E_UNKNOWN_PARM" => "Unknown parameter %s");

    my $message = sprintf $messageTags{$tag}, @args;
    print "\n$type: $message\n";
}

######################################################################
#::Tail
######################################################################

1;
